/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

#include <lz2a.h>


int log2up(int v)
{
	int c;
	c=0;
	while(v>1)
	{
		c++;
		v++;
		v>>=1;
	}
	return(c);
}

int log2upe(int v)
{
	int c;
	c=0;
	while(v>1)
	{
		c++;
		v+=3;
		v>>=2;
	}
	return(c<<1);
}

int log2upq(int v)
{
	int c;
	c=0;
	while(v>1)
	{
		c++;
		v++;
		v>>=1;
	}
	return((c+3)&(~3));
}

int log256up(int v)
{
	int c;
	c=0;
	while(v>255)
	{
		c++;
		v+=255;
		v>>=8;
	}
	return(c);
}

int log256up2(int v)
{
	int c;
	c=0;
	while(v)
	{
		c++;
		v>>=8;
	}
	return(c);
}

unsigned int PDRLELZ_DataAdler32(void *buf, int sz, unsigned int lcrc)
{
	byte *s;
	int i, c, s1, s2;

	s=buf;
	s1=lcrc&0xFFFF;
	s2=(lcrc>>16)&0xFFFF;
	for(i=0; i<sz; i++)
	{
		s1=(s1+(s[i]))%65521;
		s2=(s1+s2)%65521;
	}
	return((s2<<16)+s1);
}


//statistics/sorting

void PDRLELZ_StatBuffer(int *stat, int *num, byte *buf, int sz)
{
	int i, j, c;

	memset(stat, 0, 256*sizeof(int));
	for(i=0; i<256; i++)num[i]=i;

	for(i=0; i<sz; i++)
	{
		stat[buf[i]]++;
	}
}

int PDRLELZ_QuickSortStats(int *a, int *an, int *b, int *bn, int cnt)
{
	int *c, *d;
	int *cn, *dn;
	int min, max, med;
	int i;

	if(cnt<2)
	{
		for(i=0; i<cnt; i++)
		{
			b[i]=a[i];
			bn[i]=an[i];
		}
		return(0);
	}

	min=0x7FFFFFFF;
	max=0x80000000;
	for(i=0; i<cnt; i++)
	{
		if(a[i]<min)min=a[i];
		if(a[i]>max)max=a[i];
	}
	med=min+((max-min)/2);

	c=b;
	cn=bn;
	for(i=0; i<cnt; i++)if(a[i]>med)
	{
		*c++=a[i];
		*cn++=an[i];
	}

	d=c;
	dn=cn;
	for(i=0; i<cnt; i++)if(a[i]<=med)
	{
		*d++=a[i];
		*dn++=an[i];
	}

	if(((c-b)==cnt) || ((d-c)==cnt))
	{
/*		for(i=0; i<cnt; i++)
		{
			b[i]=a[i];
			bn[i]=an[i];
		} */
		return(0);
	}

	PDRLELZ_QuickSortStats(b, bn, a, an, c-b);
	PDRLELZ_QuickSortStats(c, cn, a+(c-b), an+(c-b), d-c);

	for(i=0; i<cnt; i++)
	{
		b[i]=a[i];
		bn[i]=an[i];
	}
	return(0);
}

void PDRLELZ_SortStats(int *stat, int *num)
{
	int i, j, k, b, c, t;
	int *cstat, *cnum;

	cstat=malloc(256*sizeof(int));
	cnum=malloc(256*sizeof(int));
	PDRLELZ_QuickSortStats(stat, num, cstat, cnum, 256);

	for(i=0; i<256; i++)
	{
		stat[i]=cstat[i];
		num[i]=cnum[i];
	}
	free(cstat);
	free(cnum);

//	return(0);
}

void PDRLELZ_SortStatsCnt(int *stat, int *num, int cnt)
{
	int i, j, k, b, c, t;
	int *cstat, *cnum;

	cstat=malloc(cnt*sizeof(int));
	cnum=malloc(cnt*sizeof(int));
	PDRLELZ_QuickSortStats(stat, num, cstat, cnum, cnt);

	for(i=0; i<cnt; i++)
	{
		stat[i]=cstat[i];
		num[i]=cnum[i];
	}
	free(cstat);
	free(cnum);

//	return(0);
}
