#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

typedef unsigned char byte;

typedef struct HTAR_InfoHeader_s HTAR_InfoHeader;

struct HTAR_InfoHeader_s {
char *header_type;
char *file_name;
long file_size;

char *content_type;
char *content_encoding;
char *content_transfer_encoding;
long content_length;
};


int htar_databytes;
char *htar_method;

int encode_file(char *fname, char *tname, char *method)
{
	FILE *ifd, *ofd;
	byte *buf;
	int i;

	if(!strcmp(method, "store"))
	{
		printf("\t <stored>\n");

		ifd=fopen(fname, "rb");
		ofd=fopen(tname, "wb");

		buf=malloc(65536);
		while(!feof(ifd))
		{
			i=fread(buf, 1, 65536, ifd);
			fwrite(buf, 1, i, ofd);
		}

		free(buf);
		fclose(ifd);
		fclose(ofd);

		return(0);
	}

	if(!strcmp(method, "gzip") || !strcmp(method, "bzip2") ||
		!strcmp(method, "lz2a"))
	{
		printf("\t <%s>\n", method);

		buf=malloc(65536);
		sprintf(buf, "%s -c %s", method, fname);

		ifd=popen(buf, "rb");
		ofd=fopen(tname, "wb");

		while(!feof(ifd))
		{
			i=fread(buf, 1, 65536, ifd);
			fwrite(buf, 1, i, ofd);
		}

		free(buf);
		pclose(ifd);
		fclose(ofd);

		return(0);
#if 0
		i=fork();
		if(!i)
		{
			execlp("bzip2", "bzip2", "-d")
		}
#endif
	}

	printf("unknown method %s for %s\n", method, fname);
	return(-1);
}

int decode_file(char *tname, char *fname, char *method)
{
	FILE *ifd, *ofd;
	byte *buf;
	int i;

	if(!strcmp(method, "store"))
	{
		printf("\t <stored>\n");

		ifd=fopen(tname, "rb");
		ofd=fopen(fname, "wb");

		buf=malloc(65536);
		while(!feof(ifd))
		{
			i=fread(buf, 1, 65536, ifd);
			fwrite(buf, 1, i, ofd);
		}

		free(buf);
		fclose(ifd);
		fclose(ofd);

		return(0);
	}

	if(!strcmp(method, "gzip") || !strcmp(method, "bzip2") ||
		!strcmp(method, "lz2a"))
	{
		printf("\t <%s>\n", method);

		buf=malloc(65536);
		sprintf(buf, "%s -dc %s", method, tname);

		ifd=popen(buf, "rb");
		ofd=fopen(fname, "wb");

		while(!feof(ifd))
		{
			i=fread(buf, 1, 65536, ifd);
			fwrite(buf, 1, i, ofd);
		}

		free(buf);
		pclose(ifd);
		fclose(ofd);

		return(0);
#if 0
		i=fork();
		if(!i)
		{
			execlp("bzip2", "bzip2", "-d")
		}
#endif
	}

	printf("unknown method %s for %s\n", method, fname);
	return(-1);
}


unsigned int PDRLELZ_DataAdler32(void *buf, int sz, unsigned int lcrc)
{
	byte *s;
	int i, c, s1, s2;

	s=buf;
	s1=lcrc&0xFFFF;
	s2=(lcrc>>16)&0xFFFF;
	for(i=0; i<sz; i++)
	{
		s1=(s1+(s[i]))%65521;
		s2=(s1+s2)%65521;
	}
	return((s2<<16)+s1);
}

char *make_date(time_t tv)
{
	static char buf[64];
	struct tm *tmv;

	tmv=gmtime(&tv);

	sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d +0000",
		1900+tmv->tm_year, tmv->tm_mon+1, tmv->tm_mday,
		tmv->tm_hour, tmv->tm_min, tmv->tm_sec);

	return(buf);
}

char *gen_tempname()
{
	static char buf[256];
	static int perm=0;
	char *s, *t;

	if(!perm)perm=rand()+(rand()<<12);
	sprintf(buf, "/tmp/htar%X.tmp", perm);
	perm+=rand();

	return(buf);
}

char *get_dirbase(char *name)
{
	static char buf[256];
	char *s, *t;

	strcpy(buf, name);
	s=buf+strlen(buf);
	while((*s!='/') && (s>buf))s--;
	*s=0;

	return(buf);
}


#if 0
int emit_convheader(FILE *ofd)
{
	static byte *cth=
		"foo\x8D\x8Abar\n"
		"\x00\x10\x7F\x80\xA1\xA2\xA3\xC0\xC1\xC2\xFD\xFE\xFF\n"
		"\t\r\n"
		"\t\tbaz\n"
		"\r\r\r"
		"\n\n\n"
		"\r\n\r\n\r\n"
		"\r\n\r\n\r\n"
		"\r\n\r\n\r\n";
	static int cthsz=55;

	char buf[300];
	int i;


	fprintf(ofd, "Header-Type: ConvTest\r\n");
	fprintf(ofd, "Content-Length: %d\r\n", cthsz);
	fprintf(ofd, "Content-Adler32: 0x%08X\r\n",
		PDRLELZ_DataAdler32(cth, cthsz, 1));
	fprintf(ofd, "Content-Encoding: store\r\n");
	fprintf(ofd, "\r\n");
	fwrite(cth, 1, cthsz, ofd);

#if 0
	for(i=0; i<256; i++)buf[i]=i;

	fprintf(ofd, "Header-Type: ConvTest256\r\n");
	fprintf(ofd, "Content-Length: %d\r\n", 256);
	fprintf(ofd, "Content-Adler32: 0x%08X\r\n",
		PDRLELZ_DataAdler32(buf, 256, 1));
	fprintf(ofd, "Content-Encoding: store\r\n");
	fprintf(ofd, "\r\n");
	fwrite(buf, 1, 256, ofd);
#endif
}
#endif

int add_file(FILE *ofd, char *name)
{
	struct stat st;
	struct stat st2;
	FILE *ifd;
	byte *buf;
	char *s;
	int i;

	i=lstat(name, &st);
	if(i<0)
	{
		printf("couldn't stat %s\n", name);
		return(-1);
	}

	s=gen_tempname();
	encode_file(name, s, htar_method);

	i=lstat(s, &st2);
	if(i<0)
	{
		printf("couldn't stat tempfile %s\n", name);
		return(-1);
	}

//	ifd=fopen(name, "rb");
	ifd=fopen(s, "rb");
	if(!ifd)
	{
		printf("couldn't open %s\n", name);
		return(-1);
	}

	fprintf(ofd, "Header-Type: File\r\n");
	fprintf(ofd, "File-Name: %s\r\n", name);
	fprintf(ofd, "File-Size: %ld\r\n", st.st_size);

	fprintf(ofd, "File-ATime: %s\r\n", make_date(st.st_atime));
	fprintf(ofd, "File-MTime: %s\r\n", make_date(st.st_mtime));
	fprintf(ofd, "File-CTime: %s\r\n", make_date(st.st_ctime));

	fprintf(ofd, "Content-Length: %ld\r\n", st2.st_size);
	fprintf(ofd, "Content-Encoding: %s\r\n", htar_method);

	fprintf(ofd, "\r\n");

	buf=malloc(65536);
	i=st2.st_size;
	while(i>65536)
	{
		fread(buf, 1, 65536, ifd);
		fwrite(buf, 1, 65536, ofd);
		i-=65536;
		htar_databytes+=65536;
	}
	if(i)
	{
		fread(buf, 1, i, ifd);
		fwrite(buf, 1, i, ofd);
		htar_databytes+=i;
	}

	fclose(ifd);
	fprintf(ofd, "\r\n");
	fprintf(ofd, "\r\n");

	return(0);
}

int add_dir(FILE *ofd, char *name)
{
	char buf[256];
	struct stat st;
	DIR *dir;
	struct dirent *de;
	int i;

	i=lstat(name, &st);
	if(i<0)
	{
		printf("couldn't stat %s\n", name);
		return(-1);
	}

	dir=opendir(name);
	if(!dir)
	{
		printf("couldn't open dir %s\n", name);
		return(-1);
	}

	fprintf(ofd, "Header-Type: Directory\r\n");
	fprintf(ofd, "File-Name: %s\r\n", name);
	fprintf(ofd, "File-Size: %ld\r\n", st.st_size);

	fprintf(ofd, "File-ATime: %s\r\n", make_date(st.st_atime));
	fprintf(ofd, "File-MTime: %s\r\n", make_date(st.st_mtime));
	fprintf(ofd, "File-CTime: %s\r\n", make_date(st.st_ctime));

	fprintf(ofd, "\r\n");

	while(1)
	{
		de=readdir(dir);
		if(!de)break;

		if(!strcmp(de->d_name, "."))continue;
		if(!strcmp(de->d_name, ".."))continue;

		sprintf(buf, "%s/%s", name, de->d_name);
		add_name(ofd, buf);
	}

	closedir(dir);

	return(0);
}

int add_name(FILE *ofd, char *name)
{
	struct stat st;
	int i;

	i=lstat(name, &st);

	if(i<0)
	{
		printf("couldn't stat %s\n", name);
		return(-1);
	}

	if(S_ISREG(st.st_mode))
	{
		add_file(ofd, name);
		return(0);
	}
	if(S_ISDIR(st.st_mode))
	{
		add_dir(ofd, name);
		return(0);
	}

	printf("ignored: %s\n", name);
	return(-1);
}

int unfiltstr(char *s)
{
	char *t;

	t=s;
	while(*s)
	{
		if(*s=='\r')
		{
			s++;
			continue;
		}
		if(*s=='\n')
		{
			s++;
			break;
		}

		if(*s=='\\')
		{
			s++;
			switch(*s++)
			{
			case 'n':
				*t++='\n';
				break;
			case 'r':
				*t++='\r';
				break;
			case 't':
				*t++='\t';
				break;
			case '\\':
				*t++='\\';
				break;
			default:
				break;
			}
			continue;
		}

		*t++=*s++;
	}
	*t++=0;
}

int read_header(FILE *ifd, HTAR_InfoHeader *hdr)
{
	static char *buf;
	static char *var, *val;
	char *s, *t;
	int ihdr, bufsz, varsz, valsz;

	t=NULL;
	s=NULL;
	ihdr=0;

	bufsz=256+(rand()%128);
	varsz=64+(rand()%16);
	valsz=16384+(rand()%4096);

	buf=malloc(bufsz);
	var=malloc(varsz);
	val=malloc(valsz);

	memset(var, 0, varsz);
	memset(val, 0, valsz);

	while(!feof(ifd))
	{
		memset(buf, 0, bufsz);
		fgets(buf, bufsz-1, ifd);
		unfiltstr(buf);

		s=buf;

		if(*s=='\t')
		{
			s++;
			while(*s)*t++=*s++;
			*t=0;

			continue;
		}

		if(var[0])
		{
			if(!strcmp(var, "Header-Type"))
				hdr->header_type=strdup(val);
			if(!strcmp(var, "File-Name"))
				hdr->file_name=strdup(val);

			if(!strcmp(var, "Content-Encoding"))
				hdr->content_encoding=strdup(val);
			if(!strcmp(var, "Content-Type"))
				hdr->content_type=strdup(val);
			if(!strcmp(var, "Content-Transfer-Encoding"))
				hdr->content_transfer_encoding=strdup(val);

			if(!strcmp(var, "File-Size"))
				hdr->file_size=atoi(val);
			if(!strcmp(var, "Content-Length"))
				hdr->content_length=atoi(val);

			memset(var, 0, varsz);
			memset(val, 0, valsz);
			ihdr=1;
		}

		if(!*s)
		{
			if(ihdr)break;
			continue;
		}


		t=var;
		while(*s && *s!=':')*t++=*s++;
		*t++=0;
		if(*s==':')s++;
		if(*s==' ')s++;

		t=val;
		while(*s)*t++=*s++;
		*t=0;
	}

	free(buf);
	free(var);
	free(val);

	return(0);
}

int dump_headers(FILE *ifd)
{
	HTAR_InfoHeader hdr;

	while(!feof(ifd))
	{
		memset(&hdr, 0, sizeof(HTAR_InfoHeader));
		read_header(ifd, &hdr);

		if(hdr.file_name)printf("%s\n", hdr.file_name);
		if(hdr.content_length)fseek(ifd, hdr.content_length, 1);
	}
}

int extract_headers(FILE *ifd)
{
	HTAR_InfoHeader hdr;
	FILE *ofd, *tfd;
	byte *buf;
	char *s;
	int i;

	buf=malloc(65536);

	while(!feof(ifd))
	{
		memset(&hdr, 0, sizeof(HTAR_InfoHeader));
		read_header(ifd, &hdr);

		if(feof(ifd))break;

		if(hdr.header_type && !strcmp(hdr.header_type, "File"))
		{
			printf("%s\n", hdr.file_name);

			s=get_dirbase(hdr.file_name);
			if(*s)mkdir(s, 0);

			s=gen_tempname();
			ofd=fopen(s, "wb");
//			ofd=fopen(hdr.file_name, "wb");

			if(ofd)
			{
				i=hdr.content_length;
				while(i>=65536)
				{
					fread(buf, 1, 65536, ifd);
					fwrite(buf, 1, 65536, ofd);
					i-=65536;
				}
				fread(buf, 1, i, ifd);
				fwrite(buf, 1, i, ofd);
				fclose(ofd);
			}else
			{
				printf("fail open %s\n", hdr.file_name);
				fseek(ifd, hdr.content_length, 1);
			}

			decode_file(s, hdr.file_name, hdr.content_encoding);
		}else
		{
			fseek(ifd, hdr.content_length, 1);
		}

	}
	free(buf);
}

int help(int argc, char *argv[])
{
	printf(
"Usage: %s [options] [files]\n"
"\n"
"Options:\n"
"	-t		dump archive\n"
"	-x		extract files\n"
"	-c		create archive\n"
"	-r		append files to archive\n"
"\n"
"	-j		use bzip2 for compressing files\n"
"	-z		use gzip for compressing files\n"
"	-l		use lz2a for compressing files\n"
"\n"
"	-f <filename>	give the filename for the archive to be used\n"

"", argv[0]);

	return(0);
}

int main(int argc, char *argv[])
{
	FILE *ofd, *ifd;
	char *s;
	int i, j, cd;

	cd=0;
	ofd=NULL;
	ifd=NULL;

	htar_method="store";

	if(argc<2)
	{
		help(argc, argv);
		return(-1);
	}

	i=1;
	while(i<argc)
	{
		s=argv[i];
		j=i+1;

		if(*s=='-')
		{
			s++;
			while(*s)
			{
				if(*s=='t')cd=1;
				if(*s=='x')cd=2;
				if(*s=='c')cd=3;
				if(*s=='r')cd=4;

				if(*s=='j')htar_method="bzip2";
				if(*s=='z')htar_method="gzip";
				if(*s=='l')htar_method="lz2a";

				if(*s=='f')switch(cd)
				{
				case 0:
					printf("need give mode "
						"before -f option\n");
					break;
				case 1:
				case 2:
					ifd=fopen(argv[j], "rb");
					j++;
					break;
				case 3:
					ofd=fopen(argv[j], "wb");
					j++;
					break;
				case 4:
					ofd=fopen(argv[j], "ab");
					j++;
					break;
				default:
					printf("invalid mode for "
						"-f option\n");
					break;
				}

				s++;
			}
			i=j;
			continue;
		}

		switch(cd)
		{
		case 3:
		case 4:
			add_name(ofd, argv[i]);
			break;
		default:
			printf("can't use files in this mode\n");
			break;
		}
		i++;
	}

	switch(cd)
	{
	case 1:
		dump_headers(ifd);
		break;
	case 2:
		extract_headers(ifd);
		break;
	}

//	ofd=fopen(argv[1], "wb");

//	htar_databytes=0;
//	emit_convheader(ofd);
//	for(i=2; i<argc; i++)
//		add_name(ofd, argv[i]);

//	printf("%d %d %d%%\n", ftell(ofd), htar_databytes,
//		(100*ftell(ofd))/htar_databytes);

	if(ifd)fclose(ifd);
	if(ofd)fclose(ofd);
}
