#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef LZ2A_H
#define LZ2A_H

#ifdef __cplusplus
extern "C" {
#endif


#define INLINE	inline		//disable if inline is not liked
//#define INLINE		//no-inline version

//#define WHUFF_NOLONGLONG	//'long long' type not present
//#define WHUFF_BITIO		//a few extra functions, may cost speed

#define WHUFF_LOWER	0x00000000
#define WHUFF_UPPER	0xFFFFFFFF
#define WHUFF_NEUTRAL	0x8000
#define WHUFF_WBITS	16
#define WHUFF_WMASK	((1<<WHUFF_WBITS)-1)

#define WARITH2_BITS	8	//W: arithmetic coder bits/symbol
#define WARITH2_CTXBITS	16	//W: arithmetic coder context bits
#define WARITH2_CLAMP	256	//W: clamp value (min distance to extremes)

#define LZ2A_RUNBITS	16	//W: bits used to encode lengths
#define LZ2A_DISTBITS	32	//W: bits used to encode distances

#define LZ2A_VERSION	0x01			//W: LZ2A version code
#define LZ2A_TESTSTR	"LZ2A TEST STRING"	//W: constant 16 char string

#define RLELZ_DICTSZ	(1<<24)	//size of dictionary
#define RLELZ_DICTSZ2	(1<<24)	//size of dictionary for opbuf
#define RLELZ_RUNSZ	(1<<16)	//max len of searched runs
#define RLELZ_DDEPTH	1024	//dict depth for each hash ent
#define RLELZ_HASHSZ	(1<<24)	//size of hash
#define RLELZ_SDEPTH	256	//search depth

//#define LZ2A_USE_OPBUF	//slow and expensive, find compound matches


typedef unsigned char byte;
//typedef unsigned short ushort;

#ifndef WHUFF_NOLONGLONG
typedef unsigned long long ull;
typedef unsigned long long ulsz;
#else
typedef unsigned long ulsz;
#endif

extern byte *huff_winbuf;	//window buffer
extern int huff_winofs;	//bit offset in byte
extern int huff_outbits;	//dummy output bits

extern unsigned int whuff_rmin;		//window lower range
extern unsigned int whuff_rmax;		//window upper range
extern unsigned int whuff_rval;		//window decode value

extern unsigned short warith2_divtab[65536];	//division table
extern unsigned int warith2_divtab2[65536];	//division table (32 bit)
extern int warith2_ctx;
extern unsigned char *warith2_model;	//pairs of counts

extern int warith2_ctxbits;
extern int warith2_ctxmask;

extern unsigned char *lz2a_lmodel;
extern unsigned char *lz2a_dmodel;

extern int lz2a_sdepth;
extern int lz2a_dictsz;
extern int lz2a_dictsz2;
extern int lz2a_runbits;
extern int lz2a_distbits;

extern char *lz2a_teststr;
extern int lz2a_verbose;

#ifdef __cplusplus
}
#endif

#endif
