/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

/*
	Basic Buffer-centric encoder/decoder
*/

#include <lz2a.h>


int PDRLELZ_CompLZ2A_Header(byte *in, byte *out, int sz)
{
	byte *cs, *ct, *ce;
	unsigned int crc;
	int i;

//	fprintf(stderr, "hello\n");

	lz2a_distbits=log2up(lz2a_dictsz);

	i=log2up(sz);
//	fprintf(stderr, "sz bits %d\n", i);
	if(i<lz2a_distbits)
	{
		lz2a_dictsz=1<<i;
		lz2a_distbits=i;
	}

	ct=out;
	//algo magic
	*ct++='L';
	*ct++='Z';
	*ct++='2';
	*ct++='A';

	*ct++=LZ2A_VERSION;	//algo version a.b, eg: 0.1
	*ct++=0x00;		//reserved, set to 0

	//rest of header info size
	*ct++=0x00;
	*ct++=0x00;

	//low order size
	*ct++=0x01;
	*ct++=6;
	*ct++=sz&0xFF;
	*ct++=(sz>>8)&0xFF;
	*ct++=(sz>>16)&0xFF;
	*ct++=(sz>>24)&0xFF;

	//checksum
	*ct++=0x02;
	*ct++=6;
	crc=PDRLELZ_DataAdler32(in, sz, 1);
	*ct++=crc&0xFF;
	*ct++=(crc>>8)&0xFF;
	*ct++=(crc>>16)&0xFF;
	*ct++=(crc>>24)&0xFF;

	//dictionary size
	*ct++=0x03;
	*ct++=6;
	*ct++=log2up(lz2a_dictsz);
	*ct++=lz2a_runbits;
	*ct++=lz2a_distbits;
	*ct++=warith2_ctxbits;

	//other bits of info can be put here

	i=ct-out;
	out[6]=i&0xFF;
	out[7]=(i>>8)&0xFF;

	//marker for begining of encoded stream
	*ct++='S';
	*ct++='T';
	*ct++='R';
	*ct++='M';


	WHUFF_SetupEncode(ct);
	LZ2A_SetupContexts();

	for(i=0; i<16; i++)
		WArith2_EncodeSymbol(lz2a_teststr[i]);

	return(ct-out);
}

int PDRLELZ_CompLZ2A(byte *in, byte *out, int sz)
{
	int cnt[256], num[256], rnum[256];
	ushort *opbuf;

	byte *cs, *ct, *ce, *cl;
//	int **hash;
	int *hash, *hcbuf;
	ushort *hlbuf;
	int dd;

	int i, j, k, l, t;
	int v[16];
	int li, lj, lk, ll, nl;
	int bl, bn, bi, bnl;	//match run/offset
	int lbl, lbn;	//last match run/offset
	int rl, rv, rc;	//run len/value


	if(!sz)return(0);

#if 1
	hash=malloc(RLELZ_HASHSZ*sizeof(int));
	memset(hash, 0, RLELZ_HASHSZ*sizeof(int));

	hcbuf=malloc(sz*sizeof(int));
	memset(hcbuf, 0, sz*sizeof(int));
#endif

#ifdef LZ2A_USE_OPBUF
	fprintf(stderr, "build opbuf\n");

	opbuf=malloc(sz*2);

	cs=in;
	ce=in+sz;
	cl=in;
	for(i=0; i<sz; i++)
	{
		if(!(i&1023))
		{
			i=cs-in;
			j=ce-in;

			fprintf(stderr, "%d/%d %2.2f%%\r",
				i, j, (100.0*i)/j);
			fflush(stderr);
		}

//		LZ2A_LookupString(in, cs, ce, hash, hcbuf,
//			&bl, &bn, lz2a_dictsz2, RLELZ_RUNSZ-1);
		LZ2A_LookupString(in, cs, ce, hash, hcbuf,
			&bl, &bn, lz2a_dictsz2, (1<<lz2a_runbits)-1);

		LZ2A_UpdateHashChar(in, cs, ce, hash, hcbuf);

		opbuf[i]=bl;
		cs++;
	}
	fprintf(stderr, "\n");
#endif

	PDRLELZ_StatBuffer(cnt, num, in, sz);
	PDRLELZ_SortStats(cnt, num);
	for(i=0; i<256; i++)rnum[num[i]]=i;
	for(i=0; i<16; i++)v[i]=num[0xF0+i];

//	v0=num[255];	//escape
//	v1=num[254];	//LZ run
//	v2=num[253];	//EOM

//	fprintf(stderr, "encode\n");

	//version dependent header
	//16 reserved codes
	for(i=0; i<16; i++)
		WArith2_EncodeSymbol(v[i]);

	//marker to flag start of data
	//also verifies that the arithmetic decoder is working
//	WArith2_EncodeSymbol('d');
//	WArith2_EncodeSymbol('a');
//	WArith2_EncodeSymbol('t');
//	WArith2_EncodeSymbol('a');


	cs=in;
	ce=in+sz;
	cl=in;
	t=1024;
	rc=0;

	while(1)
	{
		if(cs>=ce)break;

		if(!t)
		{
			i=cs-in;
			j=ce-in;
			k=huff_winbuf-out;

			fprintf(stderr, "%d/%d %2.2f%% %2.2f%%\r",
				i, j, (100.0*i)/j, (100.0*k)/i);
			fflush(stderr);
			t=1024;
		}else t--;

		//try looking up a string
		if((cs+5)<ce)
		{
			bl=0;
			bn=0;
			bi=0;
			bnl=0;

#ifdef LZ2A_USE_OPBUF
			for(i=0; i<8; i++)
			{
				k=opbuf[cs-in+i];
				if(k<=6)continue;

				for(j=0; j<8; j++)
				{
					l=opbuf[cs-in+k+j];
					if((k+l-i-j)>bl)
					{
						bl=k+l-i-j;
						bi=i;
					}
				}
			}
#endif

			LZ2A_LookupString(in, cs, ce, hash, hcbuf,
				&bl, &bn, &bi,
				lz2a_dictsz, (1<<lz2a_runbits)-1);

			while(bi--)
			{
				LZ2A_UpdateHashChar(in, cs, ce, hash, hcbuf);
				if(rnum[*cs]>=0xF0)
					WArith2_EncodeSymbol(num[255]);
				WArith2_EncodeSymbol(*cs++);
			}

			if(bl>5)
//			if(bl>=3)
//			if(bl)
			{
				lj=((cs-in)-bn);
//				if(!rc)fprintf(stderr, "%d %d %d\n",
//					cs-in, bl, lj);
				rc++;

				WArith2_EncodeSymbol(num[254]);
				LZ2A_EncodeReference(bl, lj);

				LZ2A_UpdateHash(in, cs, ce, hash, hcbuf,
					bl, bn);
				cs+=bl;
//				warith2_ctx=(cs[-2]<<8)+cs[-1];
				continue;
			}
		}
		LZ2A_UpdateHashChar(in, cs, ce, hash, hcbuf);

		if(rnum[*cs]>=0xF0)
			WArith2_EncodeSymbol(num[255]);
		WArith2_EncodeSymbol(*cs++);
	}
	WArith2_EncodeSymbol(num[253]);

	WArith2_EncodeSymbol('E');
	WArith2_EncodeSymbol('O');
	WArith2_EncodeSymbol('M');
	WArith2_EncodeSymbol(' ');

	WHUFF_FlushWBits();

	fprintf(stderr, "\n");
	fprintf(stderr, "%d %d\n", cs-in, ce-in);

	if(huff_winofs>0)huff_winbuf++;
	i=huff_winbuf-out;

	return(i);
}

int PDRLELZ_DecompLZ2A_Header(byte *in, byte *out, int sz,
	int *rsz, unsigned long *rcrc)
{
	byte *cs, *ct, *ce, *s2;
	int i, j, k;

	cs=in;
	if((cs[0]!='L') || (cs[1]!='Z') || (cs[2]!='2') || (cs[3]!='A'))
	{
		fprintf(stderr, "Invalid algo magic\n");
		return(-1);
	}

	if(cs[4]!=LZ2A_VERSION)
	{
		fprintf(stderr, "Wrong LZ2A version\n");
		return(-1);
	}

	if(cs[5])
	{
		fprintf(stderr, "Reserved value!=0\n");
		return(-1);
	}

	k=cs[6]+(cs[7]<<8);
	if((k<8) || (k>32767))
	{
		fprintf(stderr, "Invalid Header Size %d\n", k);
		return(-1);
	}


	cs=in+8;
	while(cs<(in+k))
	{
		if(cs[0]==0x01)
		{
			i=cs[2]+(cs[3]<<8)+(cs[4]<<16)+(cs[5]<<24);

			fprintf(stderr, "Size %d\n", i);
			if(rsz)*rsz=i;
			cs+=cs[1];
			continue;
		}
		if(cs[0]==0x02)
		{
			i=cs[2]+(cs[3]<<8)+(cs[4]<<16)+(cs[5]<<24);
			fprintf(stderr, "CRC %08X\n", i);

			if(rcrc)*rcrc=i;
			cs+=cs[1];
			continue;
		}

		if(cs[0]==0x03)
		{
			i=1<<cs[2];
			fprintf(stderr, "Window %d (%d bit)\n", i, cs[2]);
			lz2a_dictsz=i;

			if(cs[1]>4)
			{
				fprintf(stderr, "Run Bits %d\n", cs[3]);
				fprintf(stderr, "Dist Bits %d\n", cs[4]);
				fprintf(stderr, "Context Bits %d\n", cs[5]);
				lz2a_runbits=cs[3];
				lz2a_distbits=cs[4];
				warith2_ctxbits=cs[5];
				warith2_ctxmask=(1<<warith2_ctxbits)-1;
			}

			cs+=cs[1];
			continue;
		}

		cs+=cs[1];
	}

	cs=in+k;


	if((cs[0]!='S') || (cs[1]!='T') || (cs[2]!='R') || (cs[3]!='M'))
	{
		fprintf(stderr, "Invalid stream magic\n");
		return(-1);
	}
	cs+=4;

	WHUFF_SetupDecode(cs);
	LZ2A_SetupContexts();

	for(i=0; i<16; i++)
	{
		j=WArith2_DecodeSymbol();

		if(j!=lz2a_teststr[i])
		{
			fprintf(stderr, "Invalid test string"
				"(possible encoding/decoding issue)\n", j);
			return(-1);
		}
	}

	return(cs-in);
}

int PDRLELZ_DecompLZ2A(byte *out, int osz)
{
	byte num[256];

	int i, j, k, l, ll;
	byte *ct, *s2;

	if(!osz)return(0);

	ct=out;

	for(i=0; i<256; i++)
		num[i]=0;
	for(i=0; i<16; i++)
	{
		j=WArith2_DecodeSymbol();
		num[j]=i+0xF0;
	}

	while((ct-out)<osz)
	{
		j=WArith2_DecodeSymbol();
		i=num[j];
		if(i)
		{
			if(i==0xFF)
			{
				j=WArith2_DecodeSymbol();
				*ct++=j;
				continue;
			}
			if(i==253)
			{
				break;
			}

			if(i==0xFE)
			{
				LZ2A_DecodeReference(&l, &k);
//				fprintf(stderr, "run %d %d %d\n",
//					ct-out, l, k);

				ll=l;
				while(k<l)
				{
#if 0
					j=memcmp(ct, ct-k, k);
					if(j)
					{
						fprintf(stderr,
							"mismatch %d %d %d\n",
							ct-out, l, k);
						break;
					}
#endif
					memcpy(ct, ct-k, k);

					ct+=k;
					l-=k;
				}

#if 0
				j=memcmp(ct, ct-k, l);
				if(j)
				{
					fprintf(stderr, "mismatch %d %d %d\n",
						ct-out, l, k);
					break;
				}
#endif
				memcpy(ct, ct-k, l);

				ct+=l;
//				warith2_ctx=(ct[-2]<<8)+ct[-1];

				continue;
			}

			fprintf(stderr, "Invalid control code\n");
//			return(-1);
		}

#if 0
		if(*ct!=j)
		{
			fprintf(stderr, "mismatch %d\n", ct-out);
			break;
		}
#endif
		*ct++=j;
		continue;
	}

#if 1
	j=WArith2_DecodeSymbol();
	if(num[j]==253)j=WArith2_DecodeSymbol();
	j+=WArith2_DecodeSymbol()<<8;
	j+=WArith2_DecodeSymbol()<<16;
	j+=WArith2_DecodeSymbol()<<24;
	if(j!=('E'+('O'<<8)+('M'<<16)+(' '<<24)))
	{
		fprintf(stderr, "Missing 'EOM ' tag %08X"
			"(possible encoding/decoding issue)\n", j);
	}
#endif

	return(ct-out);
}


int PDRLELZ_CompLZ2A_Raw(byte *in, byte *out, int sz)
{
	fprintf(stderr, "hello raw\n");

	WHUFF_SetupEncode(out);
	LZ2A_SetupContexts();
	return(0);
}

int PDRLELZ_DecompLZ2A_Raw(byte *in, byte *out, int sz)
{
	WHUFF_SetupDecode(in);
	LZ2A_SetupContexts();
	return(0);
}
