/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

#include <lz2a.h>


unsigned char *lz2a_lmodel;
unsigned char *lz2a_dmodel;

int lz2a_sdepth=RLELZ_SDEPTH;
int lz2a_dictsz=RLELZ_DICTSZ;
int lz2a_dictsz2=RLELZ_DICTSZ2;
int lz2a_runbits=16;
int lz2a_distbits=32;

char *lz2a_teststr=LZ2A_TESTSTR;
int lz2a_verbose=1;


int LZ2A_SetupContexts()
{
	int i, j, k;


	for(i=0; i<256; i++)
		for(j=0; j<256; j++)
	{
		k=((i+1)<<16)/(i+j+2);
		warith2_divtab[i+(j<<8)]=k;
		warith2_divtab2[i+(j<<8)]=k<<16;
	}


	if(!lz2a_lmodel)lz2a_lmodel=malloc(256*32*2*sizeof(unsigned char));
	if(!lz2a_dmodel)lz2a_dmodel=malloc(256*64*2*sizeof(unsigned char));

	for(i=0; i<256*32*2; i++)
		lz2a_lmodel[i]=0;
	for(i=0; i<256*64*2; i++)
		lz2a_dmodel[i]=0;

//	j=1<<WARITH2_CTXBITS;
	j=1<<warith2_ctxbits;
	if(!warith2_model)warith2_model=malloc(j*2*sizeof(unsigned char));
	for(i=0; i<j*2; i++)warith2_model[i]=0;
	warith2_ctx=0;
}

int LZ2A_EncodeReference(int l, int k)
{
	int i, j, v, t;

//	WArith2_EncodeSymbol(256);

	v=0;
//	i=LZ2A_RUNBITS;
	i=lz2a_runbits;
//	for(i=0; i<LZ2A_RUNBITS; i++)
	while(i--)
	{
//		t=(l>>((LZ2A_RUNBITS-1)-i))&1;
		t=(l>>i)&1;

//		j=WArith2_Predict(lz2a_lmodel, i);
//		WArith2_Update(lz2a_lmodel, i, t);

		j=WArith2_Predict(lz2a_lmodel, (v<<4)+i);
		WArith2_Update(lz2a_lmodel, (v<<4)+i, t);
		v=((v<<1)|t)&0xFF;

//		j=WHUFF_NEUTRAL;
		WHUFF_OutputBit(t, j);
	}

	v=0;
//	for(i=0; i<LZ2A_DISTBITS; i++)
//	i=LZ2A_DISTBITS;
	i=lz2a_distbits;
	while(i--)
	{
//		t=(k>>((LZ2A_DISTBITS-1)-i))&1;
		t=(k>>i)&1;

//		j=WArith2_Predict(lz2a_dmodel, i);
//		WArith2_Update(lz2a_dmodel, i, t);

		j=WArith2_Predict(lz2a_dmodel, (v<<5)+i);
		WArith2_Update(lz2a_dmodel, (v<<5)+i, t);
		v=((v<<1)|t)&0xFF;

//		j=WHUFF_NEUTRAL;
		WHUFF_OutputBit(t, j);
	}

	return(i);
}

void LZ2A_DecodeReference(int *rl, int *rk)
{
	int i, j, k, l, v, t;

//	WArith2_EncodeSymbol(256);

	v=0;
	l=0;
//	for(i=0; i<LZ2A_RUNBITS; i++)
	for(i=0; i<lz2a_runbits; i++)
	{
		t=WHUFF_InputModelBit(lz2a_lmodel, (v<<4)+i);

		l=(l<<1)|t;
		v=((v<<1)|t)&0xFF;
	}

	v=0;
	k=0;
//	for(i=0; i<LZ2A_DISTBITS; i++)
	for(i=0; i<lz2a_distbits; i++)
	{
		t=WHUFF_InputModelBit(lz2a_dmodel, (v<<5)+i);
		k=(k<<1)|t;
		v=((v<<1)|t)&0xFF;
	}

	*rl=l;
	*rk=k;
}


//LZ data
void LZ2A_LookupString_Inner(byte *in, byte *cs, byte *ce,
	int *hash, int *hcbuf,
	int *rbl, int *rbn, int dsz, int rsz)
{
	int d, lm, cn;
	int li, lj, lk, ll;
	int bl, bn;
	byte *s, *s1, *t;

	int ci;	//current index for hcbuf

//	li=cs[0]^(cs[1]<<5)^(cs[2]<<10)^(cs[2]>>6);
//	li=cs[0]^(cs[1]<<5)^(cs[2]<<10)^(cs[0]<<15)^(cs[2]>>4);
//	li=cs[0]^(cs[1]<<8)^(cs[2]<<16);
//	li=cs[0]^(cs[1]<<5)^(cs[2]<<10)^(cs[3]<<15)^(cs[3]>>4);

//	li=cs[0]^(cs[1]<<8)^(cs[2]<<16)^cs[3]^(cs[4]<<8)^(cs[5]<<16);
//	li=cs[0]^(cs[1]<<8)^(cs[2]<<16)^(cs[3]<<4)^(cs[4]<<12)^(cs[5]<<20);
//	li=cs[0]^(cs[1]<<8)^(cs[2]<<16)^(cs[2]>>4)^(cs[3]<<4)^(cs[4]<<12);
	li=cs[0]^(cs[1]<<4)^(cs[2]<<8)^(cs[3]<<12)^(cs[4]<<16)^(cs[5]<<20);
//	li=cs[5]^(cs[4]<<4)^(cs[3]<<8)^(cs[2]<<12)^(cs[1]<<16)^(cs[0]<<20);
	li&=RLELZ_HASHSZ-1;
	ci=hash[li];

	lm=ce-cs;
	lm=(lm<rsz)?lm:rsz;

	bn=0;
	bl=0;

	ll=lz2a_sdepth;
	while(ll--)
	{
		s=in+ci;
		s1=s;
		d=cs-s;

		if(d>=dsz)break;
		if(s>=cs)
		{
			if(!ci)break;
			ci=hcbuf[ci];
			continue;
		}

		t=cs;

		lk=lm;
		while((*s++==*t++) && lk--);
		lk=(s-s1)-1;

		if(lk>bl)
		{
			bn=ci;
			bl=lk;
		}

		if(!ci)break;
		ci=hcbuf[ci];
	}

	*rbl=bl;
	*rbn=bn;
}

void LZ2A_LookupString(byte *in, byte *cs, byte *ce,
	int *hash, int *hcbuf,
	int *rbl, int *rbn, int *rbi, int dsz, int rsz)
{
	int i, j, k;
	int bl, bn, bw, bi;
	int cw, cl0, cn0, cl1, cn1;

	bl=0;
	bn=0;
	bi=0;
	bw=0;

	LZ2A_LookupString_Inner(in, cs, ce, hash, hcbuf,
		&bl, &bn, dsz, rsz);

#if 0
	for(i=0; i<4; i++)
	{
		LZ2A_LookupString_Inner(in, cs+i, ce, hash, hcbuf,
			&cl0, &cn0, dsz, rsz);

		cw=cl0-i;

		if(cw>bw)
		{
			bl=cl0;
			bn=cn0;
			bi=i;
			bw=cw;
		}

#if 0
		for(j=0; j<8; j++)
		{
			LZ2A_LookupString_Inner(in, cs+i+cl0+j,
				ce, hash, hcbuf,
				&cl1, &cn1, dsz, rsz);
			cw=cl0+cl1-i-j;

			if(cw>bw)
			{
				bl=cl0;
				bn=cn0;
				bi=i;
				bw=cw;
			}
		}
#endif
	}
#endif

	*rbl=bl;
	*rbn=bn;
	*rbi=bi;
}

void LZ2A_UpdateHashChar(byte *in, byte *cs, byte *ce,
	int *hash, int *hcbuf)
{
	int i, j;

//	i=cs[0]^(cs[1]<<5)^(cs[2]<<10)^(cs[2]>>6);
//	i=cs[0]^(cs[1]<<5)^(cs[2]<<10)^(cs[0]<<15)^(cs[2]>>4);
//	i=cs[0]^(cs[1]<<8)^(cs[2]<<16);

//	i=cs[0]^(cs[1]<<8)^(cs[2]<<16)^(cs[3]<<4)^(cs[4]<<12)^(cs[5]<<20);
//	i=cs[0]^(cs[1]<<8)^(cs[2]<<16)^(cs[2]>>4)^(cs[3]<<4)^(cs[4]<<12);
	i=cs[0]^(cs[1]<<4)^(cs[2]<<8)^(cs[3]<<12)^(cs[4]<<16)^(cs[5]<<20);
//	i=cs[5]^(cs[4]<<4)^(cs[3]<<8)^(cs[2]<<12)^(cs[1]<<16)^(cs[0]<<20);
	i&=RLELZ_HASHSZ-1;

	j=cs-in;
	hcbuf[j]=hash[i];
	hash[i]=j;
}

void LZ2A_UpdateHash(byte *in, byte *cs, byte *ce,
	int *hash, int *hcbuf,
	int bl, int bn)
{
	int i;

	for(i=0; i<bl; i++)LZ2A_UpdateHashChar(in, cs+i, ce, hash, hcbuf);
}



/*
	Attempt: Combine LZ77 and LZW style string matching
	Status: failure
*/

void LZ2A_LookupStringLZW(byte *in, byte *cs, byte *ce,
	int *hash, int *hcbuf, ushort *hlbuf,
	int *rbl, int *rbn, int dsz, int rsz)
{
	int d, lm, cn;
	int li, lj, lk, ll;
	int bl, bn;
	byte *s, *s1, *t;
	int ci;	//current index for hcbuf/hlbuf

	lm=ce-cs;
	lm=(lm<rsz)?lm:rsz;
	bn=0;
	bl=0;

//	li=cs[0]^(cs[1]<<8)^(cs[2]<<16);
	li=cs[0]^(cs[1]<<8);
	li&=RLELZ_HASHSZ-1;
//	li=0;

	ci=hash[li];

//	ll=lz2a_sdepth;
	ll=-1;
	while(ll--)
	{
		lj=hlbuf[ci<<1];
		s=in+ci-hlbuf[(ci<<1)|1];
		d=cs-s;

		if(d>=dsz)break;
		if(s>=cs)
		{
			if(!ci)break;
			ci=hcbuf[ci];
			continue;
		}

		t=cs;

		lk=lj;
		while((*s++==*t++) && lk--);

		if(t==cs)
		{
			if(!ci)break;
			ci=hcbuf[ci];
			continue;
		}

#if 1
		if(lk)
		{
			if(!ci)break;
			ci=hcbuf[ci];
			continue;
		}
#endif

		if(lj>bl)
		{
			bn=ci-hlbuf[(ci<<1)|1];
			bl=lj;
		}

		if(!ci)break;
		ci=hcbuf[ci];
	}

	*rbl=bl;
	*rbn=bn;
}

void LZ2A_UpdateHashLZW(byte *in, byte *cs, byte *ce, byte *cl,
	int *hash, int *hcbuf, ushort *hlbuf,
	int bl)
{
	int i, j, k;

	i=cl[0]^(cl[1]<<8);
	i&=RLELZ_HASHSZ-1;
//	i=0;

	j=cs-in;
	k=(cs-cl)+bl;
	if(k>65535)k=65535;

	hcbuf[j]=hash[i];
	hlbuf[(j<<1)|0]=k;
	hlbuf[(j<<1)|1]=cs-cl;
	hash[i]=j;
}
