/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

/*
	A frontend centered around using buffers for encoding/decoding.
*/

#include <lz2a.h>
#include <sys/time.h>


int Sys_TimeDUS()
{
	struct timeval tp;
	static int secbase=0, usecbase=0;
	int delta; 

	gettimeofday(&tp, NULL);

	if(!secbase)
	{
		secbase=tp.tv_sec;
		usecbase=tp.tv_usec;
	}

	delta=((tp.tv_sec-secbase)*1000000)+(tp.tv_usec-usecbase);
	secbase=tp.tv_sec;
	usecbase=tp.tv_usec;

	return(delta);
}

int help(int argc, char *argv[])
{
	int i, j;

	fprintf(stderr,
		"%s: [options] infile [outfile]\n"
		"\n"
		"Buffer-centric BGB-LZ2A encoder/decoder\n"
		"\n"
		"Options:\n"
		"\t-c\tcompress\n"
		"\t-d\tdecompress\n"
		"\t-r\tencode/decode raw stream\n"
		"\t-h\tthis help message\n"
		"\n"
		"\t-Sd <limit>\tdefine search depth (default %d)\n"
		"\t\tthis effects how many possible matches are checked\n"
		"\t\tincreasing this potentially increases compression at\n"
		"\t\tthe cost of making compression take longer\n"
		"\n"
		"\t-Szd <bits>\tdefine dictionary bits (default %d)\n"
		"\t\tthis may effect how much memory may be "
			"needed for\n"
		"\t\tbuffers (both ends), but may also help ratio\n"
		"\n"
		"\t-Szl <value>\tsize limit for raw decodes (default %d)\n"
		"\n"
		"\tnote: with raw streams these options need to be matched\n"
		"\tbetween encoding and decoding:\n"
		"\t-Cb <bits>\tdefine context bits (default %d)\n"
		"\t-Rb <bits>\tdefine run bits (default %d)\n"
		"\t-Db <bits>\tdefine distance bits (default %d)\n"
		"\n"
		"\n"
#if 0
		"\tSo <limit>\tprint input bytes usable to have "
			"output <= limit\n"
		"\tIl <limit>\tonly process limit bytes from infile\n"
		"\tOl <limit>\tonly produce limit bytes of output\n"
		"\t\tlimits do not effect the compression applied, but "
			"instead\n"
			"\t\tserve to behave as if the input had been "
			"truncated.\n"
		"\n"
#endif
		"", argv[0], lz2a_sdepth, log2up(lz2a_dictsz), (1<<24),
		warith2_ctxbits, lz2a_runbits, lz2a_distbits);

#if 0
	i=log2up(lz2a_dictsz);
	j=(1<<(i+1));		//window
	j+=(1<<(i+2));		//encode buffer
	j+=(1<<(i+3));		//hash chain buffer
	j+=RLELZ_HASHSZ*4;	//main hash

	j+=1<<(WARITH2_CTXBITS+1);	//byte context
	j+=1<<(8+4);			//length context
	j+=1<<(8+5);			//distance context
	j+=65536*2;		//division tables (16 bit)
	j+=65536*4;		//division tables (32 bit)

	fprintf(stderr, "encode approx mem use estimate %dMB\n", j>>20);


	i=log2up(lz2a_dictsz);
	j=(1<<(i+1));		//window
	j+=(1<<(i+1));		//encoded data

	j+=1<<(WARITH2_CTXBITS+1);	//byte context
	j+=1<<(8+4);			//length context
	j+=1<<(8+5);			//distance context
	j+=65536*2;		//division tables (16 bit)
	j+=65536*4;		//division tables (32 bit)

	fprintf(stderr, "decode approx mem use estimate %dMB\n", j>>20);
#endif

	return(0);
}

int main(int argc, char *argv[])
{
	FILE *ifd, *ofd;
	byte *buf, *buf2, *t;
	char *infn, *outfn;
	char *s;
	int i, j, k, l;
	int cd, szl, iszl, oszl, sd, raw;
	unsigned int crc, crc2;

#if 0
//	buf=malloc(1<<27);
//	buf2=malloc(1<<27);

	if((argc>1) && strcmp(argv[1], "-"))
		fd=fopen(argv[1], "rb");
		else fd=stdin;

	if((argc>2) && strcmp(argv[2], "-"))
		ofd=fopen(argv[2], "wb");
		else ofd=NULL;

	dc=0;
#endif

	if(argc<2)
	{
		help(argc, argv);
		exit(-1);
	}

	infn=NULL;
	outfn=NULL;
	cd=0;
	raw=0;	//headered stream
	szl=0;	//context dependent size limit
	iszl=0;	//input size limit
	oszl=0;	//output size limit

	i=1;
	while(i<argc)
	{
		if((argv[i][0]=='-') && argv[i][1])
		{
			j=i+1;

			s=argv[i]+1;
			while(*s)
			{
				if((s[0]=='S') && (s[1]=='d'))
				{
					lz2a_sdepth=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='S') && (s[1]=='z') && (s[2]=='d'))
				{
					lz2a_dictsz=1<<atoi(argv[j]);
					j++;
					s+=3;
					continue;
				}

				if((s[0]=='C') && (s[1]=='b'))
				{
					warith2_ctxbits=atoi(argv[j]);
					warith2_ctxmask=
						(1<<warith2_ctxbits)-1;
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='R') && (s[1]=='b'))
				{
					lz2a_runbits=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='D') && (s[1]=='b'))
				{
					lz2a_distbits=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}


				if((s[0]=='S') && (s[1]=='z') && (s[1]=='l'))
				{
					szl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}

				if((s[0]=='S') && (s[1]=='o'))
				{
					szl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}

				if((s[0]=='I') && (s[1]=='l'))
				{
					iszl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='O') && (s[1]=='l'))
				{
					oszl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}

				if(*s=='h')
				{
					help(argc, argv);
					exit(-1);
				}

				if(*s=='r')raw=1;

				if(*s=='c')cd=0;
				if(*s=='d')cd=1;
				s++;
			}
			i=j;
			continue;
		}

		if(!infn)
		{
			infn=argv[i];
		}else
		{
			if(!outfn)
			{
				outfn=argv[i];
			}else
			{
				printf("%s: Too Many files\n",
					argv[0]);
				help(argc, argv);
				exit(-1);
			}
		}
		i++;
	}

	Sys_TimeDUS();

//	if((argc>1) && strcmp(argv[1], "-"))
//		fd=fopen(argv[1], "rb");
//		else fd=stdin;

	ifd=NULL;
	if(infn && strcmp(infn, "-"))
	{
		ifd=fopen(infn, "rb");
	}else if(infn && !strcmp(infn, "-"))ifd=stdin;

	ofd=NULL;
	if(outfn && strcmp(outfn, "-"))
	{
		ofd=fopen(outfn, "wb");
	}else if(outfn && !strcmp(outfn, "-"))ofd=stdout;

	if(!ifd)
	{
		fprintf(stderr, "%s: Couldn't open input %s\n",
			argv[0], infn);
		exit(-1);
	}
	if(outfn && !ofd)
	{
		fprintf(stderr, "%s: Couldn't open output %s\n",
			argv[0], outfn);
		exit(-1);
	}


	if(ifd)
	{
		fseek(ifd, 0, 2);
		k=ftell(ifd);
		fseek(ifd, 0, 0);

		buf=malloc(k*2);
		fread(buf, 1, k, ifd);

		fclose(ifd);
	}

	if(!cd)
	{
		if(iszl)
		{
			if(iszl)k=(k<iszl)?k:iszl;
		}
		if(oszl)
		{
//			i=PDRLELZ_CompSizeLZ2(buf, k, oszl);
			k=(k<i)?k:i;
		}
		if(szl)k=(k<szl)?k:szl;

		crc=PDRLELZ_DataAdler32(buf, k, 1);
		fprintf(stderr, "compress %08X\n", crc);

		Sys_TimeDUS();

		buf2=malloc(k+(k/2));

		if(!raw)
		{
//			fprintf(stderr, "hello 1\n");
			PDRLELZ_CompLZ2A_Header(buf, buf2, k);
		}else PDRLELZ_CompLZ2A_Raw(buf, buf2, k);
		i=PDRLELZ_CompLZ2A(buf, buf2, k);
//		i=HUFF_EncodeData(buf, buf2, k);
		j=Sys_TimeDUS();
		fprintf(stderr, "%dB/s %fs\n\n",
			(int)(k/((double)j/1000000.0)),
			(double)j/1000000.0);

		j=k?k:1;
		fprintf(stderr, "%d in, %d out, %2.2f%%\n",
			k, i, (i*100.0)/j);

		if(ofd)
		{
			fwrite(buf2, 1, i, ofd);
			fclose(ofd);
		}

		Sys_TimeDUS();
		if(!raw)j=PDRLELZ_DecompLZ2A_Header(buf2, buf, i, NULL, NULL);
			else j=PDRLELZ_DecompLZ2A_Raw(buf2, buf, i);
		if(j<0)return(j);

		j=PDRLELZ_DecompLZ2A(buf, k);
		if(k!=j)fprintf(stderr, "%d!=%d\n", k, j);

		l=Sys_TimeDUS();
		fprintf(stderr, "%dB/s %fs\n\n",
			(int)(j/((double)l/1000000.0)),
			(double)l/1000000.0);

		crc2=PDRLELZ_DataAdler32(buf, j, 1);
		if(crc!=crc2)fprintf(stderr, "%08X!=%08X\n", crc, crc2);
	}else if(cd==1)
	{
		fprintf(stderr, "decompress\n");

		if(!raw)
		{
			j=PDRLELZ_DecompLZ2A_Header(buf, buf2, k, &i, &crc);
			if(j<0)return(j);
		}else
		{
			PDRLELZ_DecompLZ2A_Raw(buf2, buf, k);

			i=szl;
			if(!i)i=(1<<24);
			crc=0;
		}

//		i=PDRLELZ_DecompSizeLZ2(buf, k);
		buf2=malloc(i*2);


		Sys_TimeDUS();

		j=PDRLELZ_DecompLZ2A(buf2, i);
		if(i!=j)fprintf(stderr, "%d!=%d\n", i, j);

		j=Sys_TimeDUS();
		fprintf(stderr, "%dB/s %fs\n\n",
			(int)(i/((double)j/1000000.0)),
			(double)j/1000000.0);

		fprintf(stderr, "%d in, %d out, %2.2f%%\n",
			k, i, (i*100.0)/(k+1));

		if(crc)
		{
			crc2=PDRLELZ_DataAdler32(buf2, i, 1);
			if(crc!=crc2)
				fprintf(stderr, "crc %08X!=%08X\n",
					crc, crc2);
		}

		if(ofd)
		{
			fwrite(buf2, 1, i, ofd);
			fclose(ofd);
		}
	}

	return(0);
}
