/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

/*
	A file-centric frontend.
*/

#include <lz2a.h>
#include <sys/time.h>


int Sys_TimeDUS()
{
	struct timeval tp;
	static int secbase=0, usecbase=0;
	int delta; 

	gettimeofday(&tp, NULL);

	if(!secbase)
	{
		secbase=tp.tv_sec;
		usecbase=tp.tv_usec;
	}

	delta=((tp.tv_sec-secbase)*1000000)+(tp.tv_usec-usecbase);
	secbase=tp.tv_sec;
	usecbase=tp.tv_usec;

	return(delta);
}

int help(int argc, char *argv[])
{
	int i, j;

	fprintf(stderr,
		"%s: [options] infile outfile\n\n"
		"Options:\n"
		"\t-c\tcompress\n"
		"\t-d\tdecompress\n"
		"\t-h\tthis help message\n"
		"\n"
		"\t-Sd <limit>\tdefine search depth (default %d)\n"
		"\t\tthis effects how many possible matches are checked\n"
		"\t\tincreasing this potentially increases compression at\n"
		"\t\tthe cost of making compression take longer\n"
		"\n"
		"\t-Szd <bits>\tdefine dictionary bits (default %d)\n"
		"\t\tthis will effect how much memory may be "
			"needed for\n"
		"\t\tbuffers (both ends), but may also help ratio\n"
		"\t-Cb <bits>\tdefine context bits (default %d)\n"
		"\n"
#if 0
		"\t-So <limit>\tprint input bytes usable to have "
			"output <= limit\n"
		"\t-Il <limit>\tonly process limit bytes from infile\n"
		"\t-Ol <limit>\tonly produce limit bytes of output\n"
		"\t\tlimits do not effect the compression applied, but "
			"instead\n"
			"\t\tserve to behave as if the input had been "
			"truncated.\n"
		"\n"
#endif
		"", argv[0], lz2a_sdepth, log2up(lz2a_dictsz),
		warith2_ctxbits);

	i=log2up(lz2a_dictsz);
	j=(1<<(i+1));		//window
	j+=(1<<(i+2));		//encode buffer
	j+=(1<<(i+3));		//hash chain buffer
	j+=RLELZ_HASHSZ*4;	//main hash

	j+=1<<(warith2_ctxbits+1);	//byte context
	j+=1<<(8+4);			//length context
	j+=1<<(8+5);			//distance context
	j+=65536*2;		//division tables (16 bit)
	j+=65536*4;		//division tables (32 bit)

	fprintf(stderr, "encode approx mem use estimate %dMB\n", j>>20);


	i=log2up(lz2a_dictsz);
	j=(1<<(i+1));		//window
	j+=(1<<(i+1));		//encoded data

	j+=1<<(warith2_ctxbits+1);	//byte context
	j+=1<<(8+4);			//length context
	j+=1<<(8+5);			//distance context
	j+=65536*2;		//division tables (16 bit)
	j+=65536*4;		//division tables (32 bit)

	fprintf(stderr, "decode approx mem use estimate %dMB\n", j>>20);

	return(0);
}

int main(int argc, char *argv[])
{
	FILE *ifd, *ofd;
	byte *buf, *buf2, *t;
	char *infn, *outfn;
	char *s;
	int i, j, k, l;
	int cd, szl, iszl, oszl, sd;
	unsigned long crc, crc2;


	if(argc<2)
	{
		help(argc, argv);
		exit(-1);
	}

	infn=NULL;
	outfn=NULL;

	ifd=NULL;
	ofd=NULL;

	cd=0;
	szl=0;	//context dependent size limit
	iszl=0;	//input size limit
	oszl=0;	//output size limit

	i=1;
	while(i<argc)
	{
		if((argv[i][0]=='-') && argv[i][1])
		{
			j=i+1;

			s=argv[i]+1;
			while(*s)
			{
				if((s[0]=='S') && (s[1]=='d'))
				{
					lz2a_sdepth=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='S') && (s[1]=='z') && (s[2]=='d'))
				{
					lz2a_dictsz=1<<atoi(argv[j]);
					j++;
					s+=3;
					continue;
				}
				if((s[0]=='C') && (s[1]=='b'))
				{
					warith2_ctxbits=atoi(argv[j]);
					warith2_ctxmask=
						(1<<warith2_ctxbits)-1;
					j++;
					s+=2;
					continue;
				}


				if((s[0]=='S') && (s[1]=='o'))
				{
					szl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}

				if((s[0]=='I') && (s[1]=='l'))
				{
					iszl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}
				if((s[0]=='O') && (s[1]=='l'))
				{
					oszl=atoi(argv[j]);
					j++;
					s+=2;
					continue;
				}

				if(*s=='h')
				{
					help(argc, argv);
					exit(-1);
				}

				if(*s=='c')cd=0;
				if(*s=='d')cd=1;
				s++;
			}
			i=j;
			continue;
		}

		if(!infn)
		{
			infn=argv[i];
		}else
		{
			if(!outfn)
			{
				outfn=argv[i];
			}else
			{
				printf("%s: Too Many files\n",
					argv[0]);
				help(argc, argv);
				exit(-1);
			}
		}
		i++;
	}

	Sys_TimeDUS();

	if(infn && strcmp(infn, "-"))
	{
		ifd=fopen(infn, "rb");
	}else if(infn && !strcmp(infn, "-"))ifd=stdin;

	if(outfn && strcmp(outfn, "-"))
	{
		ofd=fopen(outfn, "wb");
	}else if(outfn && !strcmp(outfn, "-"))ofd=stdout;

	if(!ifd)
	{
		fprintf(stderr, "%s: Couldn't open input %s\n",
			argv[0], infn);
		exit(-1);
	}
	if(outfn && !ofd)
	{
		fprintf(stderr, "%s: Couldn't open output %s\n",
			argv[0], outfn);
		exit(-1);
	}


	if(ifd && ofd && !cd)
	{
		LZ2A_EncodeFile(ifd, ofd);
		return(0);
	}

	if(ifd && ofd && (cd==1))
	{
		LZ2A_DecodeFile(ifd, ofd);
		return(0);
	}

	fprintf(stderr, "%s: Nothing to do\n", argv[0]);
	help(argc, argv);

	return(-1);
}
