/*
	Copyright (C) 2005  Brendan G Bohannon (aka: cr88192)
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
	
	Email: cr88192@hotmail.com
*/

#include <lz2a.h>

byte *huff_winbuf;	//window buffer
int huff_winofs;	//bit offset in byte
int huff_outbits;	//dummy output bits

unsigned int whuff_rmin;		//window lower range
unsigned int whuff_rmax;		//window upper range
unsigned int whuff_rval;		//window decode value


//bit io

#ifdef WHUFF_BITIO
int HUFF_InputBit()
{
	int i;

//	i=*huff_winbuf&(1<<huff_winofs);
	i=(*huff_winbuf>>(7-huff_winofs))&1;
	huff_winofs++;

	if(huff_winofs>=8)
	{
		huff_winbuf++;
		huff_winofs-=8;
	}
	return(i);
}

void HUFF_OutputBit(int i)
{
	if(huff_winbuf)
	{
		if(i)*huff_winbuf|=(128>>huff_winofs);
			else *huff_winbuf&=~(128>>huff_winofs);

		huff_winofs++;

		if(huff_winofs>=8)
		{
			huff_winbuf++;
			huff_winofs-=8;
		}
	}else huff_outbits++;
}

int HUFF_InputByte()
{
	int i;

	i=(huff_winbuf[0]<<huff_winofs)|
		(huff_winbuf[1]>>(8-huff_winofs));
	i&=0xFF;

	huff_winbuf++;

	return(i);
}

void HUFF_OutputByte(int i)
{
	if(huff_winbuf)
	{
		huff_winbuf[0]&=~(0xFF>>huff_winofs);
		huff_winbuf[1]&=~(0xFF<<(8-huff_winofs));
		huff_winbuf[0]|=(i>>huff_winofs);
		huff_winbuf[1]|=(i<<(8-huff_winofs));
		huff_winbuf++;
	}else huff_outbits+=8;
}
#endif

#ifndef WHUFF_BITIO
INLINE int HUFF_InputByte()
{
	return(*huff_winbuf++);
}

INLINE void HUFF_OutputByte(int i)
{
	*huff_winbuf++=i;
}
#endif


//arithmetic coder core

void WHUFF_OutputBit(int i, unsigned int w)
{
	unsigned int r, r2, v;
	int j;

	r=whuff_rmax-whuff_rmin;
#ifdef WHUFF_NOLONGLONG
	v=whuff_rmin+(r>>WHUFF_WBITS)*w;
	v+=((r&WHUFF_WMASK)*w)>>WHUFF_WBITS;
#else
	v=whuff_rmin+((((ull)r)*((ull)(w<<(32-WHUFF_WBITS))))>>32);
#endif

	if(i)whuff_rmin=v+1;
	else whuff_rmax=v;

	//while bits
	while((whuff_rmin&0xFF000000)==(whuff_rmax&0xFF000000))
	{
		HUFF_OutputByte(whuff_rmin>>24);
		whuff_rmin<<=8;
		whuff_rmax<<=8;
		whuff_rmax|=255;
	}
}

int WHUFF_InputBit(unsigned int w)
{
	unsigned int r, r2, v, i;

	r=whuff_rmax-whuff_rmin;
#ifdef WHUFF_NOLONGLONG
	v=whuff_rmin+(r>>WHUFF_WBITS)*w;
	v+=((r&WHUFF_WMASK)*w)>>WHUFF_WBITS;
#else
	v=whuff_rmin+((((ull)r)*((ull)(w<<(32-WHUFF_WBITS))))>>32);
#endif

	i=(whuff_rval>v);
	if(i)whuff_rmin=v+1;
	else whuff_rmax=v;

	//while bits
	while((whuff_rmin&0xFF000000)==(whuff_rmax&0xFF000000))
	{
		whuff_rmin<<=8;
		whuff_rmax<<=8;
		whuff_rmax|=255;

		whuff_rval<<=8;
		whuff_rval|=HUFF_InputByte();
	}

	return(i);
}

int WHUFF_InputModelBit(unsigned char *model, int ctx)
{
	unsigned int r, v, i, w;

//	w=warith2_divtab2[((unsigned short *)model)[ctx]];
	r=whuff_rmax-whuff_rmin;
//	v=whuff_rmin+(((ull)r*(ull)w)>>32);

#ifdef WHUFF_NOLONGLONG
	w=warith2_divtab[((unsigned short *)model)[ctx]];
	v=whuff_rmin+(r>>WHUFF_WBITS)*w;
	v+=((r&WHUFF_WMASK)*w)>>WHUFF_WBITS;
#else
	w=warith2_divtab2[((unsigned short *)model)[ctx]];
	v=whuff_rmin+(((ull)r*(ull)w)>>32);
#endif

	i=(whuff_rval>v);
	if(i)
	{
		whuff_rmin=v+1;

		if((++model[(ctx<<1)|1])&0x80)
		{
			model[(ctx<<1)|0]>>=1;
			model[(ctx<<1)|1]>>=1;
		}
	}else
	{
		whuff_rmax=v;

		if((++model[ctx<<1])&0x80)
		{
			model[(ctx<<1)|0]>>=1;
			model[(ctx<<1)|1]>>=1;
		}
	}

	//while bits
//	while((whuff_rmin&0xFF000000)==(whuff_rmax&0xFF000000))
	while(!((whuff_rmin^whuff_rmax)&0xFF000000))
	{
		whuff_rmin<<=8;
		whuff_rmax<<=8;
		whuff_rmax|=255;

		whuff_rval<<=8;
		whuff_rval|=HUFF_InputByte();
	}

	return(i);
}

void WHUFF_FlushWBits()
{
	while((whuff_rmin!=WHUFF_LOWER) || (whuff_rmax!=WHUFF_UPPER))
	{
		HUFF_OutputByte(whuff_rmin>>24);

		whuff_rmin<<=8;
		whuff_rmax<<=8;
		whuff_rmax|=255;
	}
}


int WHUFF_SetupEncode(byte *out)
{
	huff_outbits=0;
	huff_winbuf=out;
	huff_winofs=0;

	whuff_rmin=WHUFF_LOWER;
	whuff_rmax=WHUFF_UPPER;
}

int WHUFF_SetupDecode(byte *in)
{
	int i;

	huff_outbits=0;
	huff_winbuf=in;
	huff_winofs=0;

	whuff_rmin=WHUFF_LOWER;
	whuff_rmax=WHUFF_UPPER;

	whuff_rval=WHUFF_LOWER;
	for(i=0; i<4; i++)
		whuff_rval=(whuff_rval<<8)|HUFF_InputByte();
}
